/*
 *  this file is part of sfdroid
 *  Copyright (C) 2015, Franz-Josef Haider <f_haider@gmx.at>
 *  based on harmattandroid by Thomas Perl
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include <iostream>
#include <algorithm>
#include <chrono>

#include <poll.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <sys/stat.h>
#include <grp.h>
#include "sfconnection.h"
#include "utility.h"

using namespace std;
extern int  event_fd;


hw_module_t *gralloc_module(nullptr);

static int gralloc1_release_implies_delete = 0;
static GRALLOC1_PFN_CREATE_DESCRIPTOR gralloc1_create_descriptor = NULL;
static GRALLOC1_PFN_DESTROY_DESCRIPTOR gralloc1_destroy_descriptor = NULL;
static GRALLOC1_PFN_SET_CONSUMER_USAGE gralloc1_set_consumer_usage = NULL;
static GRALLOC1_PFN_SET_DIMENSIONS gralloc1_set_dimensions = NULL;
static GRALLOC1_PFN_SET_FORMAT gralloc1_set_format = NULL;
static GRALLOC1_PFN_SET_PRODUCER_USAGE gralloc1_set_producer_usage = NULL;
static GRALLOC1_PFN_GET_BACKING_STORE gralloc1_get_backing_store = NULL;
static GRALLOC1_PFN_GET_CONSUMER_USAGE gralloc1_get_consumer_usage = NULL;
static GRALLOC1_PFN_GET_DIMENSIONS gralloc1_get_dimensions = NULL;
static GRALLOC1_PFN_GET_FORMAT gralloc1_get_format = NULL;
static GRALLOC1_PFN_GET_PRODUCER_USAGE gralloc1_get_producer_usage = NULL;
static GRALLOC1_PFN_GET_STRIDE gralloc1_get_stride = NULL;
static GRALLOC1_PFN_ALLOCATE gralloc1_allocate = NULL;
static GRALLOC1_PFN_RETAIN gralloc1_retain = NULL;
static GRALLOC1_PFN_RELEASE gralloc1_release = NULL;
static GRALLOC1_PFN_GET_NUM_FLEX_PLANES gralloc1_get_num_flex_planes = NULL;
GRALLOC1_PFN_LOCK gralloc1_lock = NULL;
static GRALLOC1_PFN_LOCK_FLEX gralloc1_lock_flex = NULL;
GRALLOC1_PFN_UNLOCK gralloc1_unlock = NULL;
#ifdef GRALLOC1_PFN_SET_LAYER_COUNT
static GRALLOC1_PFN_SET_LAYER_COUNT gralloc1_set_layer_count = NULL;
static GRALLOC1_PFN_GET_LAYER_COUNT gralloc1_get_layer_count = NULL;
#endif

gralloc1_device_t *gralloc1_device = NULL;
int  DUMMY_RENDER_TIMEOUT_MS1[2] = {DUMMY_RENDER_TIMEOUT_MS,DUMMY_RENDER_TIMEOUT_MS};

static void gralloc1_init(void)
{
    uint32_t count = 0;
    gralloc1_device->getCapabilities(gralloc1_device, &count, NULL);

    if (count >= 1) {
        int32_t i;
        int32_t *gralloc1_capabilities = (int32_t*)malloc(sizeof(int32_t) * count);

        gralloc1_device->getCapabilities(gralloc1_device, &count, gralloc1_capabilities);

        // currently the only one that affects us/interests us is release imply delete.
        for (i = 0; i < count; i++) {
#if ANDROID_VERSION_MAJOR >= 8
            if (gralloc1_capabilities[i] == GRALLOC1_CAPABILITY_RELEASE_IMPLY_DELETE) {
                gralloc1_release_implies_delete = 1;
            }
#endif
        }
		cout << "gralloc1_release_implies_delete " << gralloc1_release_implies_delete << endl;
        free(gralloc1_capabilities);
    }

    gralloc1_create_descriptor = (GRALLOC1_PFN_CREATE_DESCRIPTOR)gralloc1_device->getFunction(gralloc1_device, GRALLOC1_FUNCTION_CREATE_DESCRIPTOR);
    gralloc1_destroy_descriptor = (GRALLOC1_PFN_DESTROY_DESCRIPTOR)gralloc1_device->getFunction(gralloc1_device, GRALLOC1_FUNCTION_DESTROY_DESCRIPTOR);
    gralloc1_set_consumer_usage = (GRALLOC1_PFN_SET_CONSUMER_USAGE)gralloc1_device->getFunction(gralloc1_device, GRALLOC1_FUNCTION_SET_CONSUMER_USAGE);
    gralloc1_set_dimensions = (GRALLOC1_PFN_SET_DIMENSIONS)gralloc1_device->getFunction(gralloc1_device, GRALLOC1_FUNCTION_SET_DIMENSIONS);
    gralloc1_set_format = (GRALLOC1_PFN_SET_FORMAT)gralloc1_device->getFunction(gralloc1_device, GRALLOC1_FUNCTION_SET_FORMAT);
    gralloc1_set_producer_usage = (GRALLOC1_PFN_SET_PRODUCER_USAGE)gralloc1_device->getFunction(gralloc1_device, GRALLOC1_FUNCTION_SET_PRODUCER_USAGE);
    gralloc1_get_backing_store = (GRALLOC1_PFN_GET_BACKING_STORE)gralloc1_device->getFunction(gralloc1_device, GRALLOC1_FUNCTION_GET_BACKING_STORE);
    gralloc1_get_consumer_usage = (GRALLOC1_PFN_GET_CONSUMER_USAGE)gralloc1_device->getFunction(gralloc1_device, GRALLOC1_FUNCTION_GET_CONSUMER_USAGE);
    gralloc1_get_dimensions = (GRALLOC1_PFN_GET_DIMENSIONS)gralloc1_device->getFunction(gralloc1_device, GRALLOC1_FUNCTION_GET_DIMENSIONS);
    gralloc1_get_format = (GRALLOC1_PFN_GET_FORMAT)gralloc1_device->getFunction(gralloc1_device, GRALLOC1_FUNCTION_GET_FORMAT);
    gralloc1_get_producer_usage = (GRALLOC1_PFN_GET_PRODUCER_USAGE)gralloc1_device->getFunction(gralloc1_device, GRALLOC1_FUNCTION_GET_PRODUCER_USAGE);
    gralloc1_get_stride = (GRALLOC1_PFN_GET_STRIDE)gralloc1_device->getFunction(gralloc1_device, GRALLOC1_FUNCTION_GET_STRIDE);
    gralloc1_allocate = (GRALLOC1_PFN_ALLOCATE)gralloc1_device->getFunction(gralloc1_device, GRALLOC1_FUNCTION_ALLOCATE);
    gralloc1_retain = (GRALLOC1_PFN_RETAIN)gralloc1_device->getFunction(gralloc1_device, GRALLOC1_FUNCTION_RETAIN);
    gralloc1_release = (GRALLOC1_PFN_RELEASE)gralloc1_device->getFunction(gralloc1_device, GRALLOC1_FUNCTION_RELEASE);
    gralloc1_get_num_flex_planes = (GRALLOC1_PFN_GET_NUM_FLEX_PLANES)gralloc1_device->getFunction(gralloc1_device, GRALLOC1_FUNCTION_GET_NUM_FLEX_PLANES);
    gralloc1_lock = (GRALLOC1_PFN_LOCK)gralloc1_device->getFunction(gralloc1_device, GRALLOC1_FUNCTION_LOCK);
    gralloc1_lock_flex = (GRALLOC1_PFN_LOCK_FLEX)gralloc1_device->getFunction(gralloc1_device, GRALLOC1_FUNCTION_LOCK_FLEX);
    gralloc1_unlock = (GRALLOC1_PFN_UNLOCK)gralloc1_device->getFunction(gralloc1_device, GRALLOC1_FUNCTION_UNLOCK);
#ifdef GRALLOC1_PFN_SET_LAYER_COUNT
    gralloc1_set_layer_count = (GRALLOC1_PFN_SET_LAYER_COUNT)gralloc1_device->getFunction(gralloc1_device, GRALLOC1_FUNCTION_SET_LAYER_COUNT);
    gralloc1_get_layer_count = (GRALLOC1_PFN_GET_LAYER_COUNT)gralloc1_device->getFunction(gralloc1_device, GRALLOC1_FUNCTION_GET_LAYER_COUNT);
#endif
}


int sfconnection_t::init()
{
    int err = 0;
    struct sockaddr_un addr;

    fd_pass_socket = socket(AF_UNIX, SOCK_STREAM, 0);
    if(fd_pass_socket < 0)
    {
        cerr << "failed to create socket: " << strerror(errno) << endl;
        err = 1;
        goto quit;
    }

    memset(&addr, 0, sizeof(addr));
    addr.sun_family = AF_UNIX;
    strncpy(addr.sun_path, SHAREBUFFER_HANDLE_FILE, sizeof(addr.sun_path)-1);

    unlink(SHAREBUFFER_HANDLE_FILE);

    if(bind(fd_pass_socket, (struct sockaddr*)&addr, sizeof(addr)) < 0)
    {
        cerr << "failed to bind socket" << SHAREBUFFER_HANDLE_FILE << ": " << strerror(errno) << endl;
        err = 2;
        goto quit;
    }

#if DEBUG
    cout << "listening on " << SHAREBUFFER_HANDLE_FILE << endl;
#endif
    if(listen(fd_pass_socket, 5) < 0)
    {
        cerr << "failed to listen on socket " << SHAREBUFFER_HANDLE_FILE << ": " << strerror(errno) << endl;
        err = 3;
        goto quit;
    }
#if DEBUG
    cout << "loading gralloc module" << endl;
#endif
    if(hw_get_module(GRALLOC_HARDWARE_MODULE_ID, (const hw_module_t**)&gralloc_module) != 0)
    {
        cerr << "failed to open " << GRALLOC_HARDWARE_MODULE_ID << " module" << endl;
        err = 4;
        goto quit;
    }
    if ((gralloc1_open(gralloc_module, &gralloc1_device) == 0) && (gralloc1_device != NULL)) {
        // success
        gralloc1_init();
        //version = 1;
        //atexit(hybris_gralloc_deinitialize);
    }

    
    chmod(SHAREBUFFER_HANDLE_FILE, S_IRWXU|S_IRWXG|S_IRWXO);

 quit:
    return err;
}

void dummy_f(android_native_base_t *base)
{
}

int sfconnection_t::wait_for_buffer(int &timedout, bool &is_not_a_buffer,int &display_id,bool &iswait)
{
    int err = 0;
    int r;
    int gerr;
    char buf[32];
    ANativeWindowBuffer *buffer = nullptr;
    native_handle_t *handle = nullptr;
    int registered = 0;
    struct pollfd fdp[10];
    int nb_fdp = 0;
    int result;
    int fd_index = -1;
    std::vector<int> fd_clienti;
    timedout = 0;

    fd_clienti.push_back(fd_client[0]);
#if DEBUG
    cout << "waiting for notification size "<< fd_clienti.size() << endl;
#endif

	for (int i=0;i < fd_clienti.size();i++){
		if (fd_clienti[i] == fd_pass_socket){ //need to accept
			{
				wait_for_client();
				is_not_a_buffer = true;
			}
			continue;
		}
#if DEBUG
		cout << "waiting for notification" << endl;
#endif
  
		errno = 0;
		r = recv(fd_clienti[i], buf, 1, 0);
		if(r < 0)
		{
	
			if(errno == ETIMEDOUT || errno == EAGAIN || errno == EINTR)
			{
				timedout = 1;
				err = 0;
				goto quit;
			}

			cerr << "lost client " << strerror(errno) << endl;
			err = 1;
			goto quit;
		}
		//fprintf(stderr, "got 0x%X \n" , buf[0] );
		if(buf[0] != 0xFF && buf[0] != 0xFE && buf[0] != 0xFD  && buf[0] != 0xFC)
		{
			unsigned int index = buf[0];
#if DEBUG
			cout << "received post notification " << index << endl;
#endif
			if(index > buffers.size())
			{
				cerr << "invalid index: " << index << endl;
				err = 1;
				goto quit;
			}
			r = recv(fd_clienti[i], buf, 1, 0);
			if (r < 0){
				err = 1;
				goto quit;
			}
			iswait = (bool)buf[0];

			if (buffer_infos[index].display_id < 0 || buffer_infos[index].display_id >1){
				cerr << "invalid display_id: " << buffer_infos[index].display_id << endl;
				err = 2;
				goto quit;

			}
			//cerr << "index: " << index << endl;
			current_buffer[buffer_infos[index].display_id] = buffers[index];
			current_info[buffer_infos[index].display_id] = buffer_infos[index];
			display_id = buffer_infos[index].display_id;
			if (display_id < 2)
				is_not_a_buffer = false;
			else
				is_not_a_buffer = true;
			err = 0;
			goto quit;
		}
		if(buf[0] == 0xFC){
			cout << "buf posted0 " << endl;
			r = recv(fd_clienti[i], buf, 1, 0);
			if (r < 0){
				is_not_a_buffer = true;
				err = 0;
				goto quit;
			}
			SetEvent(event_fd);
			cout << "buf posted " << buf[0] << endl;
		}
		if(buf[0] == 0xFD)
		{
#if DEBUG
			cout << "received close event" << endl;
#endif
			sfdroid_event event;
			int isreallydestroy;
			int len;

			r = recv(fd_clienti[i], buf, 3, 0);
			if(r < 0)
			{
				if(errno == ETIMEDOUT || errno == EAGAIN || errno == EINTR)
				{
					timedout = 1;
					err = 0;
					goto quit;
				}

				cerr << "lost client " << strerror(errno) << endl;
				err = 1;
				goto quit;
			}
			display_id = (int)buf[0];
			isreallydestroy = (int)buf[1];
			len = (int)(unsigned char)buf[2];
			if (len > 0){
				int rlen = 0;
				while(rlen < len){
					r = recv(fd_clienti[i], &event.data.layer_name[rlen], len - rlen, 0);
					if(r <= 0)
					{
						if(errno == ETIMEDOUT || errno == EAGAIN || errno == EINTR)
						{
							timedout = 1;
							err = 0;
							goto quit;
						}

						cerr << "lost client " << strerror(errno) << endl;
						err = 1;
						goto quit;
					}
					rlen += r;
				}

				event.data.layer_name[len] = 0;
				//cout << "received close event "<< event.data.layer_name << endl;
				event.type = LAYER_CLOSE;
				event.display_id = display_id;
				event.isreallydestroy = isreallydestroy;
				if (multiwindow && display_id < 2){
					sfdroid_events_mutex.lock();
					sfdroid_events.push_back(event);
					sfdroid_events_mutex.unlock();
				}
			}
			is_not_a_buffer = true;
			err = 0;
			goto quit;
		}

		if(buf[0] == 0xFE)
		{
#if DEBUG
			cout << "received layer name" << endl;
#endif
			sfdroid_event event;
			int stack_id_sf = 0;
			int len;

			r = recv(fd_clienti[i], buf, 6, 0);
			if(r < 0)
			{
				if(errno == ETIMEDOUT || errno == EAGAIN || errno == EINTR)
				{
					fprintf(stderr,"name recv timeout\n");
					timedout = 1;
					err = 0;
					goto quit;
				}

				cerr << "lost client " << strerror(errno) << endl;
				err = 1;
				goto quit;
			}
			display_id = (int)buf[0];
			memcpy(&stack_id_sf,&buf[1],sizeof(int));
			len = (int)(unsigned char)buf[5];
			//cout << "received layer name0 " << display_id << " " << len<< endl;
			if (len >0){
				int rlen = 0;
				while(rlen < len){
					r = recv(fd_clienti[i], &event.data.layer_name[rlen], len - rlen, 0);
					if(r <= 0)
					{
						if(errno == ETIMEDOUT || errno == EAGAIN || errno == EINTR)
						{
							timedout = 1;
							err = 0;
							goto quit;
						}

						cerr << "lost client " << strerror(errno) << endl;
						err = 1;
						goto quit;
					}
					rlen += r;
				}
				{
					event.data.layer_name[len] = 0;
					//cout << "received layer name " << event.data.layer_name <<" " << display_id<< endl;
					event.type = LAYER_NAME;
					event.display_id = display_id;
					event.stack_id_sf = stack_id_sf;
					if (multiwindow && display_id < 2){
						sfdroid_events_mutex.lock();
						sfdroid_events.push_back(event);
						sfdroid_events_mutex.unlock();
					}
				}
			}
			is_not_a_buffer = true;
			err = 0;
			goto quit;
		}
		if(buf[0] == 0xFF)
		{
#if DEBUG
			cout << "waiting for handle" << endl;
#endif
			r = recv_native_handle(fd_clienti[i], &handle, &current_info_tmp);
			if(r < 0)
			{
				if(errno == ETIMEDOUT || errno == EAGAIN)
				{
					timedout = 1;
					err = 0;
					goto quit;
				}

				cerr << "lost client " << strerror(errno) << endl;
				err = 1;
				goto quit;
			}
			if (current_info_tmp.display_id >=2){
				is_not_a_buffer = true;
				err = 0;
				goto quit;
			}
			cerr << "before registerBuffer " << handle << " display_id: " << current_info_tmp.display_id << endl;
			current_info[current_info_tmp.display_id] = current_info_tmp;
			if (current_info_tmp.display_id >= nb_displays)
				nb_displays ++;
			{
				static int nb_buffers=0;


				cerr << "nb_buffers "<< ++nb_buffers<< endl;
			}
			gerr = gralloc1_retain(gralloc1_device, handle);
			//gerr = gralloc_module->registerBuffer(gralloc_module, handle);
			if(gerr)
			{
				cerr << "registerBuffer failed: " << strerror(-gerr) << endl;
				err = 1;
				goto quit;
			}
			buffer = new ANativeWindowBuffer();
			buffer->width = current_info_tmp.width;
			buffer->height = current_info_tmp.height;
			buffer->stride = current_info_tmp.stride;
			buffer->format = current_info_tmp.pixel_format;

			buffer->handle = handle;
			buffer->common.incRef = dummy_f;
			buffer->common.decRef = dummy_f;

			current_buffer[current_info_tmp.display_id] = buffer;

			buffers.push_back(buffer);
			buffer_infos.push_back(current_info_tmp);

#if DEBUG
			cout << "buffer info:" << endl;
			cout << "width: " << current_info_tmp.width << " height: " << current_info_tmp.height << " stride: " << current_info_tmp.stride << " pixel_format: " << current_info_tmp.pixel_format << endl;
#endif

			is_not_a_buffer = false;
		}
	quit:
		if(err != 0)
		{
			if(registered)
			{
				gralloc1_release(gralloc1_device, handle);
				if (!gralloc1_release_implies_delete) {
					native_handle_close((native_handle_t*)handle);//FIXME!!!
					native_handle_delete((native_handle_t*)handle);
				}
				//gralloc_module->unregisterBuffer(gralloc_module, handle);
			}
			// does this also make sense if layer name or layer close failed?

			fd_client.erase(std::remove_if(fd_client.begin(),fd_client.end(),[&](int fd)
			{
				return (fd == fd_clienti[i]);
			}),fd_client.end());
			close(fd_clienti[i]);
			remove_buffers();
			if(buffer) delete buffer;
			/*
			  if(handle)
			  {
			  free_handle(handle);
			  }
			*/
		}
	}
    return err;
}

void sfconnection_t::remove_buffers()
{
    for(std::vector<ANativeWindowBuffer*>::size_type i = 0;i < buffers.size();i++)
    {
        ANativeWindowBuffer *buffer = buffers[i];
        const native_handle_t *handle = buffer->handle;

		gralloc1_release(gralloc1_device, handle);
        if (!gralloc1_release_implies_delete) {
			native_handle_close((native_handle_t*)handle);
			native_handle_delete((native_handle_t*)handle);
        }

        /*gralloc_module->unregisterBuffer(gralloc_module, handle);

		  for(int i=0;i<handle->numFds;i++)
		  {
		  close(handle->data[i]);
		  }
		  free((void*)handle);
		*/
        delete buffer;
    }

    buffers.resize(0);
    buffer_infos.resize(0);
}

void sfconnection_t::send_status_and_cleanup()
{
    if(shb_fd >0)
    {
#if DEBUG
        cout << "sending status " << fd_client[shb_index] <<  endl;
#endif
        if(send_status(shb_fd, current_status) < 0)
        {
            cerr << "lost client" << endl;
			close(shb_fd);
			fd_client.erase(std::remove_if(fd_client.begin(),fd_client.end(),[&](int fd)
			{
				return (fd == shb_fd);
			}),fd_client.end());
            remove_buffers();
        }
    }
}

buffer_info_t *sfconnection_t::get_current_info(int display_id)
{
    return &current_info[display_id];
}

ANativeWindowBuffer *sfconnection_t::get_current_buffer(int display_id)
{
    return current_buffer[display_id];
}

int sfconnection_t::wait_for_client()
{
    int err = 0;
    int fd_clientl;

#if DEBUG
    cout << "waiting for client (sharebuffer module)" << endl;
#endif
    if((fd_clientl = accept(fd_pass_socket, NULL, NULL)) < 0)
    {
        cerr << "failed to accept: " << strerror(errno) << endl;
        err = 1;
        goto quit;
    }
    cout << "accept new connection" << endl;
    shb_fd = fd_clientl;//will be always last
    fd_client.push_back(fd_clientl);
    update_timeout();

 quit:
    return err;
}

void sfconnection_t::update_timeout()
{
    struct timeval timeout;
    memset(&timeout, 0, sizeof(timeout));

    if(my_have_focus)
    {
        timeout.tv_usec = SHAREBUFFER_SOCKET_TIMEOUT_US;
    }
    else
    {
        timeout.tv_sec = SHAREBUFFER_SOCKET_TIMEOUT_US;//SHAREBUFFER_SOCKET_FOCUS_LOST_TIMEOUT_S;
    }
    for (int i=0;i < fd_client.size();i++){
        if(setsockopt(fd_client[i], SOL_SOCKET, SO_RCVTIMEO, (char *)&timeout, sizeof(timeout)) < 0)
		{
    	    cerr << "failed to set timeout on sharebuffer socket: " << strerror(errno) << endl;
		}
    }
}

void sfconnection_t::thread_loop()
{
    running = true;
    int nc[2] = {1,1};
	
    while(running)
    {
        if(have_client()) update_timeout();

        if(!have_client())
        {
#if DEBUG
            cout << "waking up android" << endl;
#endif
            wakeup_android();
            if(wait_for_client() != 0)
            {
                cerr << "waiting for client failed" << endl;
            }
#if DEBUG
            else
            {
                cout << "new client" << endl;
            }
#endif
        }

        if(have_client())
        {
            int timedout = 0;
            bool is_not_a_buffer = false;
			int  display_id = 0;
			bool  iswait = 1;
            if(wait_for_buffer(timedout, is_not_a_buffer, display_id,iswait) == 0)
            {
                back_cond.notify_one();
                if(!is_not_a_buffer)
                {
                    if(!timedout)
                    {
						//cout << "iswait: " << iswait << endl;
						{
							sfdroid_event event;
							event.type = BUFFER;
							event.data.buffer.buffer = current_buffer[display_id];
							event.data.buffer.info = &current_info[display_id];
			  
							sfdroid_events_mutex.lock();
							sfdroid_events.push_back(event);
							sfdroid_events_mutex.unlock();
						}
                        unique_lock<mutex> lock(notify_mutex);
                        if(!notified)
							buffer_cond.wait(lock);
                        notified = false;
						if (iswait)
							send_status_and_cleanup();

                        // let sharebuffer know were done
						DUMMY_RENDER_TIMEOUT_MS1[display_id] = DUMMY_RENDER_TIMEOUT_MS;
						nc[display_id] = 1;
                        timeout_count[display_id] = 0;
                    }
                    else
                    {
						for (int i=0; i < nb_displays;i++){
			
							if(((timeout_count[i] + 1) * SHAREBUFFER_SOCKET_TIMEOUT_US) / 1000 >= DUMMY_RENDER_TIMEOUT_MS1[i])
							{
								if(buffers.size() > 0)
								{
                            	    sfdroid_event event;
                            	    event.type = NO_BUFFER;
                            	    event.data.buffer.buffer = current_buffer[i];
                            	    event.data.buffer.info = &current_info[i];
                            	    sfdroid_events_mutex.lock();
                            	    sfdroid_events.push_back(event);
                            	    sfdroid_events_mutex.unlock();
                            	    unique_lock<mutex> lock(notify_mutex);

                            	    if(!notified) buffer_cond.wait(lock);
                            	    notified = false;
								}
								if (nc[i] > 60 && nc[i] <= 260){
									DUMMY_RENDER_TIMEOUT_MS1[i] += DUMMY_RENDER_TIMEOUT_MS /10;
								}
								//cout << "timeout " << i <<" "<< DUMMY_RENDER_TIMEOUT_MS1[i] << endl;
								nc[i]++;
								timeout_count[i] = 0;
							}
							else timeout_count[i]++;
						}
                        if(my_have_focus)
                        {
#if DEBUG
                            cout << "wakeing up android" << endl;
#endif
                            wakeup_android();
                        }
                    }
                }
            }
        }

        std::this_thread::yield();
    }
    thread_exited = true;
}

void sfconnection_t::notify_buffer_done(int failed)
{
    unique_lock<mutex> lock(notify_mutex);
    current_status = failed;
    notified = true;
    buffer_cond.notify_one();
}

void sfconnection_t::start_thread()
{
    thread_exited = false;
    my_thread = std::thread(&sfconnection_t::thread_loop, this);
}

void sfconnection_t::stop_thread()
{
    running = false;
    
    for (int i=0;i < fd_client.size();i++)
		shutdown(fd_client[i], SHUT_RDWR);
    fd_client.clear();
    if(fd_pass_socket >= 0) shutdown(fd_pass_socket, SHUT_RDWR);
    // TODO
    while(!thread_exited)
    {
        usleep(100);
        buffer_cond.notify_one();
    }
    my_thread.join();
}

void sfconnection_t::wait_for_event(int timeout)
{
    unique_lock<mutex> notify_back_lock(notify_back_mutex);
    back_cond.wait_for(notify_back_lock, std::chrono::milliseconds(timeout));
}

bool sfconnection_t::have_client()
{
    return (fd_client.size() > 0);
}

void sfconnection_t::lost_focus()
{
    my_have_focus = false;
}

void sfconnection_t::gained_focus()
{
    my_have_focus = true;
}

void sfconnection_t::deinit()
{
    remove_buffers();
    if(fd_pass_socket >= 0) close(fd_pass_socket);
    for (int i=0;i < fd_client.size();i++){
		close(fd_client[i]);
    }
    fd_client.clear();
    unlink(SHAREBUFFER_HANDLE_FILE);
}

