/*
 *  this file is part of sfdroid
 *  Copyright (C) 2015, Franz-Josef Haider <f_haider@gmx.at>
 *  based on harmattandroid by Thomas Perl
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include "utility.h"
#include "sfdroid_defs.h"
#include <android-version.h>
#include <hybris/properties/properties.h>
#include <fstream>
#include <string>
#include <unistd.h>
#include <cstring>
#include <iostream>
#include <time.h>
#include <QStringList>
#include <QProcess>
#include <QDebug>

using namespace std;
QStringList proc;
int proc_nb=0;
// these should stay in the same window
bool is_blacklisted(string app)
{
    if(app.find("BootAnimation") == 0)
    {
        return false;
    }
    if(app.find("Animation") != string::npos || app.find("animation") != string::npos)
    {
        return true;
    }
    if(app.find("Background") != string::npos || app.find("background") != string::npos)
    {
        return true;
    }
#if 1
    if(app.find("Splash") != string::npos || app.find("splash") != string::npos)
    {
        return true;
    }
#endif
#if 0
    if(app.find("SurfaceView") == 0)
    {
        return true;
    }
#endif
    if(app.find("Task=") == 0 || app.find("Stack=") == 0)
    {
        return false;
    }
    if(app.find("SnapshotStartingWindow") != string::npos)
    {
        return false;
    }

    if(app.find(".") == std::string::npos)
    {
        return true;
    }

    if(app.find("android/com.android.internal.app.ResolverActivity") != std::string::npos)
    {
        return true;
    }

#if 0
    if(app.find("com.android.phone") != std::string::npos)
    {
        return true;
    }
#endif

    return false;
}

string get_app_name(char *layer_name)
{
    std::string app = layer_name,apps;
    std::string::size_type idx = app.find("SurfaceView ");
    std::string def_app = "com.android.launcher3";
    //return layer_name;
    //cout << endl << "get_app_name " << layer_name << endl;
    apps = app;
    if(app == "BootAnimation")
    {
        return def_app;
    }

    if(app.find("Android is starting") != std::string::npos)
    {
        return def_app;
    }

    if(app.find("StatusBar") != std::string::npos)
    {
        return def_app;
    }
    if(app.find("NavigationBar") != std::string::npos)
    {
        return def_app;
    }
    if(app.find("com.android.systemui.ImageWallpaper") != std::string::npos || app.find(def_app) != std::string::npos)
    {
        return def_app;
    }

    if(idx != std::string::npos)
    {
        app = app.substr(idx + 12);
	idx = app.find("- ");
	if(idx != std::string::npos){
	    app = app.substr(idx + 2);
	}
    }
    //cout <<  "get_app_name1 " << app << endl;
    idx = app.find("Starting ");
    if(idx != std::string::npos)
    {
        app = app.substr(idx + 9);
    }

    if(app == "")
    {
        app = layer_name;
    }

    idx = app.find(" ");
    if(idx != std::string::npos)
    {
        app = app.substr(0, idx);
		if (app.find(".") == std::string::npos)
			app = apps.substr(idx + 1,string::npos);
    }


    idx = app.find("/");
    if(idx != std::string::npos)
    {
        app = app.substr(0, idx);
    }
    return app;
}

void touch(const char *fname)
{
    fstream f;
    f.open(fname, ios::out);
}

void wakeup_android()
{
    return;
#if ANDROID_VERSION_MAJOR == 4 && ANDROID_VERSION_MINOR == 4
    system("/usr/bin/sfdroid_powerup &");
#else
    system("/usr/bin/sfdroid_powerup_cm12.1 &");
#endif
}

bool if_process_exist(const char *pname)
{
     QProcess builder;
     QString s;
     int     pid=0;

     builder.start("/bin/pidof", QStringList() << pname);
     
     if (!builder.waitForFinished())
      return false;
     s = builder.readAll();
     sscanf(s.toUtf8().data(),"%d",&pid);
     if (pid > 0)
      return true;
     
     return false;
}
void to_front(const char *app)
{
    char buff[5120];
	QString s;
	QStringList l = QString(app).split("_replace");

	if (!if_process_exist("system_server"))
		return;
	if (l.size() >0)
		s = l[0];
	else
		s = app;
    touch(AM_START_STILL_RUNNING_FILE);
    if(strcmp(app, "com.android.systemui") != 0)
    {
        snprintf(buff, 5120, "(/system/bin/am previous  -p %s; rm %s) &", s.toUtf8().data(), AM_START_STILL_RUNNING_FILE);
    }
    else
    {
        snprintf(buff, 5120, "(/system/bin/am previous  -p com.android.launcher3; rm %s) &", AM_START_STILL_RUNNING_FILE);
    }
    cout << "to_front " << buff << endl;
    system(buff);
}
void to_front_background(const char *app)
{
    char buff[5120];
	QString s;
	QStringList l = QString(app).split("_replace");

	if (!if_process_exist("system_server"))
		return;
	if (l.size() >0)
		s = l[0];
	else
		s = app;
    touch(AM_START_STILL_RUNNING_FILE);
    if(strcmp(app, "com.android.systemui") != 0)
    {
        snprintf(buff, 5120, "(/system/bin/am %s; rm %s) &", s.toUtf8().data(), AM_START_STILL_RUNNING_FILE);
    }
    else
    {
        snprintf(buff, 5120, "(/system/bin/am  com.android.launcher3; rm %s) &", AM_START_STILL_RUNNING_FILE);
    }
    cout << "to_front_background " << buff << endl;
    system(buff);
}

void find_running_processes()
{
	QProcess p;
	QString cmdb="am stack list|grep taskId|cut -d ' ' -f4";
	QString s;
	QStringList l;

	if (!if_process_exist("system_server"))
		return;
	p.start("/bin/sh",QStringList() << "-c" << cmdb);
    if (p.waitForFinished()){
    }
    s = p.readAll();
    qDebug() << s;
	proc = s.split('\n');	
}
void start_running_process()
{
	QProcess p, *pp;
	QString cmd="am start %1";

	if (proc_nb < proc.length() && proc[proc_nb].contains("com.android.launcher3"))
		proc_nb++;
	
	if (proc_nb >= proc.length() || proc[proc_nb]==""){
		proc_nb++;
		return;
	}
	qDebug() << "starting" << proc[proc_nb];
	pp = new QProcess;
	pp->setProcessChannelMode(QProcess::ForwardedChannels);
	pp->start("/bin/sh",QStringList() << "-c" << cmd.arg(proc[proc_nb]));
	//qDebug() << "starting result" << pp->readAll();
	if (pp->waitForStarted()){
	}
	proc_nb++;
	//usleep(1000000);
}
bool to_front_still_processing()
{
    return false;
#if 0	
    fstream f;
    f.open(AM_START_STILL_RUNNING_FILE, ios::in);
    if(f.is_open()) return true;
    return false;
#endif
}

void start_app(const char *appandactivity)
{
    char buff[5120];
    snprintf(buff, 5120, "/system/bin/am start --user 0 -n %s &", appandactivity);
    system(buff);
}

void go_home()
{
    system("/system/bin/am start --user 0 -c android.intent.category.HOME -a android.intent.action.MAIN &");
}

void stop_app(const char *appandactivity)
{
    char app[4196];
    char buff[5120];

    strncpy(app, appandactivity, sizeof(app)-1);
    char *slash = strstr(app, "/");
    if(slash != NULL) *slash = 0;
    snprintf(buff, 5120, "/system/bin/am force-stop --user 0 %s &", app);
    system(buff);
}
void block_input(int display_id,bool isblock)
{
    const char *pname="surfaceflinger.disable.input";
    char v[1024];
    int  vi = -1;

    property_get(pname,v,"-1");
    vi = atoi(v);
    if (isblock){
		if (vi == -1)
			vi = (display_id + 1);
		else {
			vi |= (display_id + 1);
		}
    }
    else {
		if (vi != -1)
			vi &= ~(display_id + 1);
		if (vi == 0)
			vi = -1;
    }
    snprintf(v,sizeof(v)-1,"%d",vi);
    property_set(pname,v);
    snprintf(v,sizeof(v)-1,"/system/bin/dumpsys input >/dev/null");
    system(v);
}
void resize_window(int display_id,int h)
{
    char v[1024];
    char s[1024];

    snprintf(s,sizeof(s)-1,"surfaceflinger.forceY.%d",display_id);
    snprintf(v,sizeof(v)-1,"%d",h);
    property_set(s,v);
    snprintf(v,sizeof(v)-1,"/system/bin/dumpsys input >/dev/null");
    system(v);
}

uint64_t get_current_time_mcs()
{
    uint64_t r;
    struct timespec tm;
    
    clock_gettime( CLOCK_MONOTONIC, &tm );
    //fprintf(stderr,"%ld %ld\n", tm.tv_sec,tm.tv_nsec);
    r = (uint64_t)(tm.tv_sec)*1000000 +  (uint64_t)tm.tv_nsec/1000;
    return r;
}
