/*
 *  this file is part of sfdroid
 *  Copyright (C) 2015, Franz-Josef Haider <f_haider@gmx.at>
 *  based on harmattandroid by Thomas Perl
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
#include <sstream>
#include <iostream>
#include <vector>
#include <algorithm>

#include "wayland_helper.h"
#include "windowmanager.h"
#include "utility.h"

using namespace std;

renderer_t* arender[2]={NULL,NULL};
renderer_t* parender[2]={NULL,NULL};
bool issplash[2]={0,0};
class alayer_t
{
    public:
	std::string name;
	int task_id;
};
alayer_t alayer[2];
int  event_fd ;

int find_slot(vector<int> &slot_to_fingerId, int fingerId)
{
    // find the slot
    vector<int>::iterator it = find(slot_to_fingerId.begin(), slot_to_fingerId.end(), fingerId);
    if(it != slot_to_fingerId.end())
    {
        return std::distance(slot_to_fingerId.begin(), it);
    }

    // find first free slot
    vector<int>::size_type i;
    for(i = 0;i < slot_to_fingerId.size();i++)
    {
        if(slot_to_fingerId[i] == -1)
        {
            slot_to_fingerId[i] = fingerId;
            return i;
        }
    }

    // no free slot found, add new one
    slot_to_fingerId.resize(slot_to_fingerId.size()+1);
    slot_to_fingerId[slot_to_fingerId.size()-1] = fingerId;

    return slot_to_fingerId.size()-1;
}

void erase_slot(vector<int> &slot_to_fingerId, int fingerId)
{
    vector<int>::iterator it = find(slot_to_fingerId.begin(), slot_to_fingerId.end(), fingerId);

    if(it != slot_to_fingerId.end())
    {
        vector<int>::size_type idx = distance(slot_to_fingerId.begin(), it);

        slot_to_fingerId[idx] = -1;

        // if we're at the end of the vector, erase unneeded elements
        if(idx == slot_to_fingerId.size()-1)
        {
            while(slot_to_fingerId[idx] == -1)
            {
                slot_to_fingerId.resize(idx);
                idx--;
            }
        }
    }
    else
    {
        cerr << "BUG: erase_slot" << endl;
    }
}

int windowmanager_t::init(sfconnection_t &sfc)
{
    sfconnection = &sfc;

    swipe_hack_dist_x = (SWIPE_HACK_PIXEL_PERCENT * wayland_helper::width) / 100;
    swipe_hack_dist_y = (SWIPE_HACK_PIXEL_PERCENT * wayland_helper::height) / 100;

#if DEBUG
    cout << "swipe hack dist (x,y): (" << swipe_hack_dist_x << "," << swipe_hack_dist_y << ")" << endl;
#endif
    event_fd = CreateEvent();
    return uinput.init(wayland_helper::width, wayland_helper::height);
}

void windowmanager_t::deinit()
{
    for(auto it=windows.begin();it!=windows.end();it++)
    {
        stop_app((*it)->get_package().c_str());
        stop_app((*it)->get_package().c_str());
        (*it)->deinit();
        delete *it;
    }
}

void windowmanager_t::take_focus(int display_id)
{
    for(auto wit = windows.begin();wit != windows.end();wit++)
    {
        if((*wit)->is_active() && !find_window_not((*wit)->app,display_id))
        {
            (*wit)->lost_focus();
            taken_focus = *wit;
        }
    }
}

void windowmanager_t::seat_handle_capabilities(void *data, struct wl_seat *seat, uint32_t caps)
{
    windowmanager_t *windowmanager = (windowmanager_t*)data;

#if DEBUG
    cout << "seat handle caps: " << caps << endl;
#endif
    if((caps & WL_SEAT_CAPABILITY_TOUCH))
    {
        windowmanager->w_touch = wl_seat_get_touch(seat);
        wl_touch_add_listener(windowmanager->w_touch, &(windowmanager->w_touch_listener), data);
    }
    if((caps & WL_SEAT_CAPABILITY_KEYBOARD))
    {
        windowmanager->w_keyboard = wl_seat_get_keyboard(seat);
        wl_keyboard_add_listener(windowmanager->w_keyboard, &(windowmanager->w_keyboard_listener), data);
    }
}
renderer_t* windowmanager_t::find_window(string app,int display_id)
{
    renderer_t *ro = NULL;
    auto it = std::find_if(windows.begin(),windows.end(),[&](renderer_t *r)
			  {
			    return (r->app == app && r->display_id == display_id);
			  });
    if (it != windows.end()){
	//cout << "found window " << app << " " << display_id <<endl;
	return *it;
    }
    else
	return NULL;

}
std::vector<renderer_t*> windowmanager_t::find_window(string app)
{
    std::vector<renderer_t*> ro;

    for(auto wit = windows.begin();wit != windows.end();wit++){
	if ((*wit)->app == app)
	    ro.push_back(*wit);
    }
    return ro;
}
renderer_t* windowmanager_t::find_window_task(int task_id,int display_id)
{
    renderer_t *ro = NULL;
    auto it = std::find_if(windows.begin(),windows.end(),[&](renderer_t *r)
	{
		return (r->task_id == task_id && r->display_id == display_id);
	});
    if (it != windows.end()){
		return *it;
    }
    else
		return NULL;

}

renderer_t* windowmanager_t::find_window_type(int display_id,int nb,int &nbo)
{
    renderer_t *ro = NULL;
    int cnt = 0;

    for (auto it=windows.begin();it!=windows.end();++it){
		if ((*it)->display_id == display_id && !ro)
			ro = *it;
		cnt ++;
    }
    nbo = cnt;
    if (cnt == nb){
		return ro;
    }
    else
		return NULL;
}


renderer_t* windowmanager_t::find_window_not(string app,int display_id)
{
    renderer_t *ro = NULL;
    auto it = std::find_if(windows.begin(),windows.end(),[&](renderer_t *r)
			  {
			    return (r->app == app && r->display_id != display_id);
			  });
    if (it != windows.end())
	return *it;
    else
	return NULL;

}
void windowmanager_t::parse_task_info(std::string layer_name, int & stack_id, int & task_id)
{
	int pos1,pos2,pos3;

	stack_id = task_id = -1;
	string ln = layer_name;
	pos1 = ln.find("##",0);
	if (pos1 != string::npos){
	    pos2 = ln.find("##",pos1 + 2);
	    if (pos2 != string::npos){
		pos3 = ln.find("##",pos2 + 2);
		if (pos3 != string::npos){
		    stack_id = atoi(ln.substr(pos1+2,pos2-1).c_str());
		    task_id = atoi(ln.substr(pos2+2,pos3-1).c_str());
		    //cout << "ids " << stack_id << " " << task_id << endl;
		}
	    }
	}
}

void windowmanager_t::handle_layer_name_event(char *layer_name,int display_id,int stack_id_sf)
{
#if DEBUG
    //if (!multiwindow)
	cout << "handle layer name event " << layer_name << " display_id " << display_id << endl;
#endif
    int stack_id = -1;
    int task_id = -1;

    if (strstr(layer_name,"Splash Screen")){
		issplash[display_id] = true;
    }
    else
		issplash[display_id] = false;
    if(is_blacklisted(layer_name))
    {
        // seems like a new app is starting... lets hope...
        // if a window that is toplevel here it will gain focus again @ gain focus again
        take_focus(display_id);
        //if(last_layer.find("Android is starting") == std::string::npos) return;
		return;
    }
    last_layer = layer_name;	
    wait_for_next_layer_name = false;

    string app = get_app_name(layer_name);
    {
		int pos1,pos2;
		string ln = layer_name;
		string s = "SnapshotStartingWindow for taskId=";
		pos1 = ln.find(s,0);
		if (pos1 != string::npos){
			pos2 = ln.find("#",pos1 + 1);
			if (pos2 != string::npos){
				int task_id=-1;
		
				task_id = atoi(ln.substr(s.length(),pos2-1).c_str());
				renderer_t *r = find_window_task(task_id,display_id);
				if (r){
					app = r->app;
				}
				else if (task_id >=0)//it means no window yet created so skip such event
					return;
			}
			else
				return;
		}
    }
	
    renderer_t *r = find_window(app,display_id);
    if(r == NULL && last_layer.find("Android is starting") != std::string::npos)
    {
        app = "com.android.systemui";
    }
    if(r)
    {
		//cout << "app " << app << " active " << r->is_active() << " d " << display_id << endl;
#if 0
		if(!r->is_active())
			r->gained_focus(display_id);
#endif
#if 0
        if(1/*!r->istop*//*!r->is_active()*/)
        {
            take_focus(display_id);
            // close it to bring it to the front...
            if(taken_focus != r)
            {
                //r->recreate();
                taken_focus = nullptr;
            }
            else
            {
                // gain focus again
                r->gained_focus(display_id);
                taken_focus = nullptr;
            }
        }
#endif
	task_id = r->task_id;
    }
    else
    {
		uint64_t ct;

		parse_task_info(std::string(layer_name),stack_id,task_id);

		if (multiwindow){//first window per display always launcher
			int cnt;

			find_window_type(display_id,0,cnt);
			if (!cnt){
				app = "com.android.launcher3";
				stack_id = -1;
				task_id = -1;
			}
		}
		ct = get_current_time_mcs();
		dtasks.erase(std::remove_if(dtasks.begin(),dtasks.end(),[&](task t)
		{
			return (ct > (t.time + 3*1000000 ));//remove if it is > 3sec
		}),dtasks.end());

		if (task_id >=0){
			auto it = std::find_if(dtasks.begin(),dtasks.end(),[&](task t)
			{
				return (t.id == task_id);
			});
			if (it != dtasks.end()){
				cout << "Task " << task_id << " was already removed" << endl;
				return;
			}
		}
		if (task_id >=0)
		{
			renderer_t *r = find_window_task(task_id,display_id);
			if (r){//we may launch subwindow(like com.settings) so no need to create new surface
				r->app = app + "_replace";
				r->gained_focus(display_id);
				alayer[display_id].name = app;
			    alayer[display_id].task_id = task_id;
#if DEBUG
				cout << "already have window with task_id " << task_id << endl;
#endif
				return;
			}
		}		
		cout << "new renderer " << display_id << " stack " << stack_id << " task " << task_id <<endl;
		take_focus(display_id);
		renderer_t *rn;
		rn = new renderer_t();
		rn->app = app;
		rn->layer_name = layer_name;
		rn->display_id = display_id;
		rn->stack_id_sf = stack_id_sf;
		rn->stack_id = stack_id;
		rn->task_id = task_id;
		rn->multiwindow = multiwindow;
		rn->init(*this);
		rn->set_package(app);
		rn->gained_focus(display_id);
		rn->event_fd = event_fd;
		windows.push_back(rn);
		taken_focus = nullptr;
		parender[display_id] = arender[display_id];
		arender[display_id] = rn;
    }
    alayer[display_id].name = app;
    alayer[display_id].task_id = task_id;
	
}

void windowmanager_t::handle_layer_close_event(char *layer_name,int display_id,int isreallydestroy)
{
    bool isexisting=false;
    bool istaskend=false;
    int  stack_id = -2;
    int  task_id = -2;
    string app;
#if 1//DEBUG
    cout << "handle layer close event " << layer_name  << " " << display_id << " isdestroy " << isreallydestroy << endl;
#endif

    if(is_blacklisted(layer_name))
    {
        take_focus(display_id);
        return;
    }
    //cout <<"close 1 " << endl;
    //Task=
    //Stack=
    for(auto wit = windows.begin();wit != windows.end();wit++){
	if ((*wit)->display_id == display_id && ((std::string)(layer_name)).find((*wit)->app) != std::string::npos){
	    app = (*wit)->app;
	    isexisting=true;
	    break;
	}
    }
    //cout <<"close 2 " << endl;
    if (!isexisting)
		app = get_app_name(layer_name);

    // why does this happen?
    cout << "close app " << app << " d " << display_id << " ir "<< isreallydestroy << endl;
#if 0
    {
	std::vector<renderer_t *> ra = find_window(app);
	for (int i=0;i < ra.size();i++){
	    cout << "close app " << app << "  active "<< ra[i]->is_active() <<" d "<< i << endl;
	    if(ra[i]->is_active())
		ra[i]->lost_focus();
	}
    }
#endif
    if(app == "com.android.systemui" || app == "com.android.launcher3")
    {
        //take_focus(display_id);
        return;
    }

    {
	int pos1,pos2;

	pos1 = string(layer_name).find("Task=");
	if (pos1 == string::npos){	
	    /*
	    parse_task_info(std::string(layer_name),stack_id,task_id);
	    if (task_id < 0 || !isreallydestroy)
	    */
		return;
	}
	else {
	    pos2 = string(layer_name).find("#");
	    task_id = atoi(string(layer_name).substr(pos1+5,(pos2!=string::npos)?pos2-1:-1).c_str());
	}
    }
    //if (!isreallydestroy)
    //	return;
    for (int i=0;i < 2;i++){//uff we receive on destroy event not per screen if app was killed
		renderer_t *r = find_window_task(task_id,i);
		if(r)
		{
			cout << "remove window "<< r->app << " d " << i << " t " << task_id << endl;
    	    take_focus(i);
			windows.erase(std::remove_if(windows.begin(),windows.end(),[&](renderer_t *ri)
			{
			    return (ri->task_id == task_id && ri->display_id == i);
			}),windows.end());
			r->deinit();

			if (arender[i] == r)
				arender[i] = NULL;
			if (parender[i] == r)
				parender[i] = NULL;
			delete r;
			if (i == 0){
				task t;

				t.id = task_id;
				t.time = get_current_time_mcs();
				dtasks.push_back(t);
			}
		}
    }
    if(windows.size() == 0)
    {
            sfdroid_event event;
            event.type = LAST_WINDOW_CLOSED;
            sfdroid_events_mutex.lock();
            sfdroid_events.push_back(event);
            sfdroid_events_mutex.unlock();
			cout << "last window closed";
    }
}

bool windowmanager_t::handle_buffer_event(ANativeWindowBuffer *buffer, buffer_info_t &info,int display_id)
{
#if DEBUG
    cout << "handle buffer event beg displayid: " << display_id << endl;
#endif
    int i=0;
    int nb;
    renderer_t *r;

#if 0
    if (issplash[display_id]){
		cout << "splash ignoring " << endl;
		return true;
    }
#endif
    r = find_window_type(display_id,1,nb);//if 1 only
    if (r){
        if(r->render_buffer(buffer, info, false) != 0)
        {
            return false;
        }
        return true;
    }
    for(auto wit = windows.begin();wit != windows.end();wit++)
    {
		if ((*wit)->display_id == display_id && arender[display_id] != NULL){
#if DEBUG
			cout << "handle buffer " << alayer[display_id].name << " " <<  arender[display_id]->app 
			<< " d " << display_id << " lt " <<  alayer[display_id].task_id << " rt " << arender[display_id]->task_id << endl;
#endif
		}
        if(/*(*wit)->is_active() &&(*wit)->istop*/
		   arender[display_id] && (*wit) == arender[display_id] && 
		   ( arender[display_id]->app == alayer[display_id].name || arender[display_id]->task_id == alayer[display_id].task_id)&&
		   /* 
			  (
			  (arender[display_id] &&  arender[display_id]->app == alayer[display_id]) ||
			  (arender[1-display_id] &&  arender[1-display_id]->app == alayer[display_id])
			  ) &&
		   */
		   (*wit)->display_id == display_id /*&& !wait_for_next_layer_name*/
		   )
        {
			//cout << "handle buffer event displayid: " << display_id  << " " << (*wit)->app << endl;
            if((*wit)->render_buffer(buffer, info, false) != 0)
            {
                return false;
            }			
            return true;
        }
    }
    if (parender[display_id]){//draw in previous top
		if (!arender[display_id]->isprevnative){
			parender[display_id]->render_buffer(buffer, info, false);
#if DEBUG
			cout << "handle buffer (fallback) event displayid: " << display_id  << " " << parender[display_id]->app << endl;
#endif
		}
		else
			cout << "handle buffer ignore prev was native " << display_id  << endl;
    }
    else {
	    if (arender[display_id])
			arender[display_id]->render_buffer(buffer, info, false);
#if DEBUG
		cout << "can not handle buffer(fallback 2) event displayid: " << display_id << " " << alayer[display_id].name << " " << (arender[display_id]?arender[display_id]->app:std::string("unknown"))  <<endl;
#endif
    }
    return false;
}

bool windowmanager_t::handle_no_buffer_event(ANativeWindowBuffer *old_buffer, buffer_info_t &info,int display_id)
{
#if DEBUG
    cout << "handle no buffer event displayid: " << display_id << endl;
#endif

    for(auto wit = windows.begin();wit != windows.end();wit++)
    {
        if(/*(*wit)->is_active() &&*/ /*(*wit)->istop &&*/ (*wit)->display_id == display_id)
        {
            //(*wit)->render_buffer(old_buffer, info, true);
            return true;
        }
    }

    return false;
}

void windowmanager_t::handle_close(struct wl_surface *surface)
{
#if DEBUG
    cout << "handle close" << endl;
#endif

    for(auto wit = windows.begin();wit != windows.end();wit++)
    {
        if((*wit)->get_surface() == surface)
        {
            stop_app((*wit)->get_package().c_str());
            (*wit)->deinit();
            delete *wit;
            windows.erase(wit);
            if(windows.size() == 0)
            {
                sfdroid_event event;
                event.type = LAST_WINDOW_CLOSED;
                sfdroid_events_mutex.lock();
                sfdroid_events.push_back(event);
                sfdroid_events_mutex.unlock();
            }
            break;
        }
    }
}

void windowmanager_t::touch_handle_down(void *data, struct wl_touch *wl_touch, uint32_t serial, uint32_t time, struct wl_surface *surface, int32_t id, wl_fixed_t w_x, wl_fixed_t w_y)
{
#if DEBUG
    cout << "handle touch down event" << endl;
#endif

    windowmanager_t *windowmanager = (windowmanager_t*)data;
    int x, y;
    int touch_x = wl_fixed_to_int(w_x);
    int touch_y = wl_fixed_to_int(w_y);
    int slot;

    return ;
    slot = find_slot(windowmanager->slot_to_fingerId, id);
    cout << "slot " << slot <<" id " <<id << endl;
    windowmanager->uinput.send_event(EV_ABS, ABS_MT_SLOT, slot);
    windowmanager->uinput.send_event(EV_ABS, ABS_MT_TRACKING_ID, id);

    x = touch_x;

    if(touch_x <= windowmanager->swipe_hack_dist_x)
    {
#if DEBUG
        cout << "swipe hack x" << endl;
#endif
        x = 0;
    }
    if(touch_x >= wayland_helper::width - windowmanager->swipe_hack_dist_x)
    {
#if DEBUG
        cout << "swipe hack x" << endl;
#endif
        x = wayland_helper::width;
    }

    y = touch_y;

    if(touch_y <= windowmanager->swipe_hack_dist_y)
    {
#if DEBUG
        cout << "swipe hack y" << endl;
#endif
        y = 0;
    }
    if(touch_y >= wayland_helper::height - windowmanager->swipe_hack_dist_y)
    {
#if DEBUG
        cout << "swipe hack y" << endl;
#endif
        y = wayland_helper::height;
    }
#if 0
    windowmanager->uinput.send_event(EV_ABS, ABS_MT_POSITION_X, x);
    windowmanager->uinput.send_event(EV_ABS, ABS_MT_POSITION_Y, y);
    windowmanager->uinput.send_event(EV_ABS, ABS_MT_PRESSURE, MAX_PRESSURE);
    windowmanager->uinput.send_event(EV_SYN, SYN_REPORT, 0);
#endif
}

void windowmanager_t::touch_handle_up(void *data, struct wl_touch *wl_touch, uint32_t serial, uint32_t time, int32_t id)
{
#if DEBUG
    cout << "handle touch up event" << endl;
#endif

    windowmanager_t *windowmanager = (windowmanager_t*)data;
    int slot;

    slot = find_slot(windowmanager->slot_to_fingerId, id);
#if 0
    windowmanager->uinput.send_event(EV_ABS, ABS_MT_SLOT, slot);
    windowmanager->uinput.send_event(EV_ABS, ABS_MT_TRACKING_ID, -1);
    windowmanager->uinput.send_event(EV_SYN, SYN_REPORT, 0);
#endif
    erase_slot(windowmanager->slot_to_fingerId, id);
}

void windowmanager_t::touch_handle_motion(void *data, struct wl_touch *wl_touch, uint32_t time, int32_t id, wl_fixed_t w_x, wl_fixed_t w_y)
{
#if DEBUG
    cout << "handle touch motion event" << endl;
#endif
    return ;
    windowmanager_t *windowmanager = (windowmanager_t*)data;
    int touch_x = wl_fixed_to_int(w_x);
    int touch_y = wl_fixed_to_int(w_y);
    int slot;

    slot = find_slot(windowmanager->slot_to_fingerId, id);
#if 0
    windowmanager->uinput.send_event(EV_ABS, ABS_MT_SLOT, slot);
    windowmanager->uinput.send_event(EV_ABS, ABS_MT_TRACKING_ID, id);
    windowmanager->uinput.send_event(EV_ABS, ABS_MT_POSITION_X, touch_x);
    windowmanager->uinput.send_event(EV_ABS, ABS_MT_POSITION_Y, touch_y);
    windowmanager->uinput.send_event(EV_ABS, ABS_MT_PRESSURE, MAX_PRESSURE);
    windowmanager->uinput.send_event(EV_SYN, SYN_REPORT, 0);
#endif
}

void windowmanager_t::touch_handle_frame(void *data, struct wl_touch *wl_touch)
{
#if DEBUG
    cout << "handle touch frame event" << endl;
#endif
}

void windowmanager_t::touch_handle_cancel(void *data, struct wl_touch *wl_touch)
{
#if DEBUG
    cout << "handle keyboard cancel event" << endl;
#endif
}

void windowmanager_t::keyboard_handle_keymap(void *data, struct wl_keyboard *wl_keyboard, uint32_t format, int32_t fd, uint32_t size)
{
#if DEBUG
    cout << "handle keyboard keymap event" << endl;
#endif
}

void windowmanager_t::keyboard_handle_enter(void *data, struct wl_keyboard *wl_keyboard, uint32_t serial, struct wl_surface *surface, struct wl_array *keys)
{
#if 1//DEBUG
    cout << "handle keyboard enter event" << endl;
#endif

    windowmanager_t *windowmanager = (windowmanager_t*)data;

    for(auto wit = windowmanager->windows.begin();wit != windowmanager->windows.end();wit++)
    {
        if((*wit)->get_surface() == surface)
        {
            if(!(*wit)->is_active())
            {
		wakeup_android();
#if 0
		windowmanager->take_focus((*wit)->display_id);//FIXME

		to_front((*wit)->get_package().c_str());
		if (multiwindow)
            	    windowmanager->wait_for_next_layer_name = true;
#endif
                (*wit)->gained_focus((*wit)->display_id);
                windowmanager->taken_focus = nullptr;
            }
            break;
        }
    }

    windowmanager->sfconnection->gained_focus();
}

void windowmanager_t::keyboard_handle_leave(void *data, struct wl_keyboard *wl_keyboard, uint32_t serial, struct wl_surface *surface)
{
#if DEBUG
    cout << "handle keyboard leave event" << endl;
#endif

    windowmanager_t *windowmanager = (windowmanager_t*)data;

    for(auto wit = windowmanager->windows.begin();wit != windowmanager->windows.end();wit++)
    {
        if((*wit)->get_surface() == surface)
        {
            if((*wit)->is_active())
            {
                (*wit)->lost_focus();
		cout << "keyboard_handle_leave" << endl;
                break;
            }
        }
    }

    windowmanager->sfconnection->lost_focus();
}

void windowmanager_t::keyboard_handle_key(void *data, struct wl_keyboard *wl_keyboard, uint32_t serial, uint32_t time, uint32_t key, uint32_t state)
{
#if DEBUG
    cout << "handle keyboard key event" << endl;
#endif
}

void windowmanager_t::keyboard_handle_modifiers(void *data, struct wl_keyboard *wl_keyboard, uint32_t serial, uint32_t mods_depressed, uint32_t mods_latched, uint32_t mods_locked, uint32_t group)
{
#if DEBUG
    cout << "handle keyboard modifier event" << endl;
#endif
}
 
void windowmanager_t::keyboard_handle_repeat_info(void *data, struct wl_keyboard *wl_keyboard, int32_t rate, int32_t delay)
{
#if DEBUG
    cout << "handle keyboard repeat info event" << endl;
#endif
}

